import numpy as np
import xarray as xr
from karabo.bound import (
    BOOL_ELEMENT,
    FLOAT_ELEMENT,
    KARABO_CLASSINFO,
    NODE_ELEMENT,
    OVERWRITE_ELEMENT,
    UINT32_ELEMENT,
    UINT64_ELEMENT,
    Epochstamp,
    Hash,
    ImageData,
    Timestamp,
    Trainstamp,
)
from .DetectorAssembler import DetectorAssembler
from .correction_addons.saturation_monitor import saturation_monitoring_schema
from ._version import version as deviceVersion


@KARABO_CLASSINFO("SaturationWarningAggregator", deviceVersion)
class SaturationWarningAggregator(DetectorAssembler):
    @staticmethod
    def expectedParameters(expected):
        saturation_monitoring_schema(expected)
        (
            OVERWRITE_ELEMENT(expected)
            .key("imageDataPath")
            .setNewDefaultValue("saturationMonitor.alarmImage")
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("imageMaskPath")
            .setNewDefaultValue("")
            .commit(),
        )

    def on_matched_data(self, tid, sources):
        my_timestamp = Timestamp(Epochstamp(), Trainstamp(tid))
        warn = False
        alarm = False
        warn_count = 0
        alarm_count = 0
        max_value = 0

        processed_any = False

        # TODO: look at image, too (DetectorAssembler functionality needed)

        image_datas, module_indices = [], []
        for source, (data, timestamp) in sources.items():
            if not data.has("saturationMonitor"):
                continue
            processed_any = True
            warn = warn or data["saturationMonitor.warning"]
            alarm = alarm or data["saturationMonitor.alarm"]
            warn_count += data["saturationMonitor.warnCount"]
            alarm_count += data["saturationMonitor.alarmCount"]
            if data["saturationMonitor.maxValue"] > max_value:
                max_value = data["saturationMonitor.maxValue"]

            # this part like DetectorAssembler
            image_data = data[self._image_data_path]
            if isinstance(image_data, ImageData):
                image_data = image_data.getData()
            # note: not using image mask, let preview handler zero out NaN
            image_datas.append(image_data.astype(np.float32, copy=False))
            module_indices.append(self._source_to_index[source])

        if not processed_any:
            self.log.WARN(
                "No sources in match had 'saturationMonitor', "
                "please check upstream configuration!"
            )

        # handle warning property updates
        update = Hash()

        if warn != self.get("saturationMonitor.warning"):
            update["saturationMonitor.warning"] = warn
        if alarm != self.get("saturationMonitor.alarm"):
            update["saturationMonitor.alarm"] = alarm

        if warn or alarm:
            update["saturationMonitor.warnCount"] = warn_count
            update["saturationMonitor.alarmCount"] = alarm_count
            update["saturationMonitor.trainId"] = tid
            update["saturationMonitor.maxValue"] = max_value

        if update:
            self.set(update)

        # handle maybe sending preview
        if self._geometry is None:
            self.log.WARN("Have not received a geometry yet, cannot show preview")
        elif warn or alarm:
            dims = ["module", "slow_scan", "fast_scan"]
            coords = {"module": module_indices}
            assembled_data, _ = self._geometry.position_modules(
                xr.DataArray(
                    data=image_datas,
                    dims=dims,
                    coords=coords,
                )
            )
            self._preview_friend.write_outputs(assembled_data, timestamp=my_timestamp)

        self.info["sent"] += 1
        self.info["trainId"] = tid
        self.rate_out.update()
