from karabo.bound import (
    DOUBLE_ELEMENT,
    INT32_ELEMENT,
    OVERWRITE_ELEMENT,
    VECTOR_UINT32_ELEMENT,
    VECTOR_DOUBLE_ELEMENT,
)
from collections import deque
import numpy as np
from .base_kernel import BaseArbiterKernel


def module_number(source):
    token = source.split("/")[-1]
    return int(token[:-8]) if token.endswith("CH0:xtdf") else None


class IntensityThreshold(BaseArbiterKernel):
    def __init__(self, name, config, host_device):
        super().__init__(name, config, host_device)
        self._key = "integratedIntensity.mean"
        self._modules = set(config.get("modules"))
        self._snr = config.get("snr")
        self._min_frames = config.get("minFrames")
        self._history = deque([], self._min_frames)

    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            OVERWRITE_ELEMENT(schema)
            .key(prefix)
            .setNewDescription(
                "This kernel selects the frames by comparing the averaged"
                "intensity to the threshold."
            )
            .commit(),

            VECTOR_UINT32_ELEMENT(schema)
            .key(f"{prefix}.modules")
            .assignmentOptional()
            .defaultValue([3, 4, 8, 15])
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.snr")
            .assignmentOptional()
            .defaultValue(3.5)
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(schema)
            .key(f"{prefix}.minFrames")
            .assignmentOptional()
            .defaultValue(100)
            .reconfigurable()
            .commit(),
        )

    @staticmethod
    def extend_output_schema(schema, name):
        (
            DOUBLE_ELEMENT(schema)
            .key(f"{name}.threshold")
            .assignmentOptional()
            .defaultValue(np.nan)
            .commit(),

            VECTOR_DOUBLE_ELEMENT(schema)
            .key(f"{name}.intensity")
            .assignmentOptional()
            .defaultValue([])
            .commit()
        )

    def reconfigure(self, config):
        if config.has("modules"):
            self._modules = set(config.get("modules"))
        if config.has("snr"):
            self._snr = config.get("snr")
        if config.has("minFrames"):
            self._min_frames = config.get("minFrames")
            self._history = deque(self._history, self._min_frames)

    def consider(self, train_id, sources, num_frames, mask, out_hash):
        if mask.sum() == 0:
            out_hash.set(f"{self._name}.threshold", 0)
            out_hash.set(f"{self._name}.intensity", [])
            return mask

        intensity = np.mean(
            [
                data[self._key][mask]
                for source, (data, _) in sources.items()
                if data.has(self._key) and (module_number(source) in self._modules)
            ],
            axis=0,
        )
        self._history.extend(intensity[-self._min_frames:])

        if num_frames < self._min_frames:
            q1, mu, q3 = np.percentile(intensity, [25, 50, 75])
        else:
            q1, mu, q3 = np.percentile(self._history, [25, 50, 75])

        sig = (q3 - q1) / 1.34896
        threshold = mu + self._snr * sig

        hits = intensity > threshold

        out_hash.set(f"{self._name}.threshold", threshold)
        out_hash.set(f"{self._name}.intensity", intensity)

        return mask & hits
