import enum

import numpy as np
from karabo.bound import (
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
)

from .. import (
    base_calcat,
    base_correction,
    base_kernel_runner,
    schemas,
    utils,
)
from ..preview_utils import FrameSelectionMode, PreviewFriend, PreviewSpec
from .._version import version as deviceVersion


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 1


class Constants(enum.Enum):
    Offset = enum.auto()


correction_steps = (("offset", CorrectionFlags.OFFSET, {Constants.Offset}),)


class DsscBaseRunner(base_kernel_runner.BaseKernelRunner):
    _correction_flag_class = CorrectionFlags
    _correction_steps = correction_steps
    num_pixels_ss = 128
    num_pixels_fs = 512

    def expected_input_shape(self, num_frames):
        return (
            num_frames,
            1,
            self.num_pixels_ss,
            self.num_pixels_fs,
        )

    def _setup_constant_buffers(self):
        self.offset_map = self._xp.empty(self._map_shape, dtype=np.float32)

    def flush_buffers(self, constants):
        if Constants.Offset in constants:
            self.offset_map.fill(0)

    def _load_constant(self, constant_type, data):
        assert constant_type is Constants.Offset
        # can have an extra dimension for some reason
        if len(data.shape) == 4:  # old format (see offsetcorrection_dssc.py)?
            data = data[..., 0]
        # shape (now): x, y, memory cell
        data = self._xp.asarray(data, dtype=np.float32).transpose()
        self.offset_map[:] = data[:self._constant_memory_cells]

    def _preview_data_views(self, raw_data, processed_data):
        return (
            raw_data[:, 0],
            processed_data,
        )


class DsscCpuRunner(DsscBaseRunner):
    _xp = np

    def _post_init(self):
        from ..kernels import dssc_cython
        self.correction_kernel = dssc_cython.correct

    def _correct(self, flags, image_data, cell_table, output):
        self.correction_kernel(
            image_data,
            cell_table,
            flags,
            self.offset_map,
            output,
        )


class DsscGpuRunner(DsscBaseRunner):
    def _pre_init(self):
        import cupy
        self._xp = cupy

    def _post_init(self):
        self.correction_kernel = self._xp.RawModule(
            code=base_kernel_runner.get_kernel_template("dssc_gpu.cu").render(
                {
                    "pixels_x": self.num_pixels_fs,
                    "pixels_y": self.num_pixels_ss,
                    "output_data_dtype": utils.np_dtype_to_c_type(
                        self._output_dtype
                    ),
                    "corr_enum": utils.enum_to_c_template(CorrectionFlags),
                }
            )
        ).get_function("correct")

    def _correct(self, flags, image_data, cell_table, output):
        num_frames = self._xp.uint16(image_data.shape[0])
        block_shape = (1, 1, 64)
        grid_shape = utils.grid_to_cover_shape_with_blocks(output.shape, block_shape)
        image_data = image_data[:, 0]
        self.correction_kernel(
            grid_shape,
            block_shape,
            (
                image_data,
                cell_table,
                num_frames,
                self._constant_memory_cells,
                flags,
                self.offset_map,
                output,
            ),
        )


class DsscCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = Constants

    @property
    def _constants_need_conditions(self):
        return {Constants.Offset: self.dark_condition}

    @staticmethod
    def add_schema(schema):
        super(DsscCalcatFriend, DsscCalcatFriend).add_schema(schema, "DSSC-Type")
        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(400)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(100)  # TODO: proper
            .commit()
        )

        base_calcat.add_status_schema_from_enum(schema, Constants)

    def dark_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        return res


@KARABO_CLASSINFO("DsscCorrection", deviceVersion)
class DsscCorrection(base_correction.BaseCorrection):
    _base_output_schema = schemas.xtdf_output_schema
    _correction_steps = correction_steps
    _calcat_friend_class = DsscCalcatFriend
    _constant_enum_class = Constants

    _preview_outputs = [
        PreviewSpec(name, frame_reduction_selection_mode=FrameSelectionMode.PULSE)
        for name in ("raw", "corrected")
    ]

    @classmethod
    def expectedParameters(cls, expected):
        cls._calcat_friend_class.add_schema(expected)
        DsscBaseRunner.add_schema(expected)
        base_correction.add_addon_nodes(expected, cls)
        PreviewFriend.add_schema(expected, cls._preview_outputs)
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(
                cls._base_output_schema(use_shmem_handles=cls._use_shmem_handles)
            )
            .commit(),
        )
        (
            # support both CPU and GPU kernels
            STRING_ELEMENT(expected)
            .key("kernelType")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(base_kernel_runner.KernelRunnerTypes.GPU.name)
            .options(
                ",".join(
                    kernel_type.name
                    for kernel_type in base_kernel_runner.KernelRunnerTypes
                )
            )
            .reconfigurable()
            .commit(),
        )

    @property
    def _kernel_runner_class(self):
        kernel_type = base_kernel_runner.KernelRunnerTypes[
            self.unsafe_get("kernelType")
        ]
        if kernel_type is base_kernel_runner.KernelRunnerTypes.CPU:
            return DsscCpuRunner
        else:
            return DsscGpuRunner
