import concurrent.futures
import enum

import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
)

from .. import (
    base_calcat,
    base_correction,
    base_kernel_runner,
    schemas,
    utils,
)
from .._version import version as deviceVersion
from ..preview_utils import PreviewFriend, PreviewSpec
from ..kernels import common_cython


class Constants(enum.Enum):
    BadPixelsDarkEPix100 = enum.auto()
    RelativeGainEPix100 = enum.auto()
    OffsetEPix100 = enum.auto()
    NoiseEPix100 = enum.auto()


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 2**1
    COMMONMODE = 2**2
    RELGAIN = 2**3
    BPMASK = 2**4


correction_steps = (
    ("offset", CorrectionFlags.OFFSET, {Constants.OffsetEPix100}),
    ("relGain", CorrectionFlags.RELGAIN, {Constants.RelativeGainEPix100}),
    ("commonMode", CorrectionFlags.COMMONMODE, {Constants.NoiseEPix100}),
    ("badPixels", CorrectionFlags.BPMASK, {Constants.BadPixelsDarkEPix100}),
)


class Epix100CalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = Constants

    @property
    def _constants_need_conditions(self):
        return {
            Constants.OffsetEPix100: self.dark_condition,
            Constants.BadPixelsDarkEPix100: self.dark_condition,
            Constants.NoiseEPix100: self.dark_condition,
            Constants.RelativeGainEPix100: self.illuminated_condition,
        }

    @staticmethod
    def add_schema(schema):
        super(Epix100CalcatFriend, Epix100CalcatFriend).add_schema(
            schema, "ePix100-Type"
        )

        # set some defaults for common parameters
        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(200)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsX")
            .setNewDefaultValue(708)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsY")
            .setNewDefaultValue(768)
            .commit(),
        )

        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.integrationTime")
            .tags("managed")
            .displayedName("Integration Time")
            .assignmentOptional()
            .defaultValue(10)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.sensorTemperature")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(264.32933824373004)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.inVacuum")
            .tags("managed")
            .displayedName("In vacuum")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.sourceEnergy")
            .tags("managed")
            .displayedName("Source Energy")
            .assignmentOptional()
            .defaultValue(8.04778)
            .reconfigurable()
            .commit(),
        )

        base_calcat.add_status_schema_from_enum(schema, Constants)

    def dark_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Integration Time"] = self._get_param("integrationTime")
        res["Sensor Temperature"] = self._get_param("sensorTemperature")
        res["In vacuum"] = self._get_param("inVacuum")

        return res

    def illuminated_condition(self):
        res = self.dark_condition()

        res["Source Energy"] = self._get_param("sourceEnergy")

        return res


class Epix100CpuRunner(base_kernel_runner.BaseKernelRunner):
    _correction_steps = correction_steps
    _correction_flag_class = CorrectionFlags
    _bad_pixel_constants = {Constants.BadPixelsDarkEPix100}
    _xp = np

    num_pixels_ss = 708
    num_pixels_fs = 768

    @classmethod
    def add_schema(cls, schema):
        super(cls, cls).add_schema(schema)
        super(cls, cls).add_bad_pixel_config(schema)
        (
            DOUBLE_ELEMENT(schema)
            .key("corrections.commonMode.noiseSigma")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(5)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("corrections.commonMode.minFrac")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0.25)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(schema)
            .key("corrections.commonMode.enableRow")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(schema)
            .key("corrections.commonMode.enableCol")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(schema)
            .key("corrections.commonMode.enableBlock")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),
        )

    def reconfigure(self, config):
        super().reconfigure(config)
        if config.has("corrections.commonMode.noiseSigma"):
            self._cm_sigma = np.float32(config["corrections.commonMode.noiseSigma"])
        if config.has("corrections.commonMode.minFrac"):
            self._cm_minfrac = np.float32(config["corrections.commonMode.minFrac"])
        if config.has("corrections.commonMode.enableRow"):
            self._cm_row = config["corrections.commonMode.enableRow"]
        if config.has("corrections.commonMode.enableCol"):
            self._cm_col = config["corrections.commonMode.enableCol"]
        if config.has("corrections.commonMode.enableBlock"):
            self._cm_block = config["corrections.commonMode.enableBlock"]

    def expected_input_shape(self, num_frames):
        assert num_frames == 1, "ePix not expected to have multiple frames"
        return (self.num_pixels_ss, self.num_pixels_fs)

    def _expected_output_shape(self, num_frames):
        return (self.num_pixels_ss, self.num_pixels_fs)

    @property
    def _map_shape(self):
        return (self.num_pixels_ss, self.num_pixels_fs)

    def _make_output_buffers(self, num_frames, flags):
        # ignore parameters
        return [
            self._xp.empty(
                (self.num_pixels_ss, self.num_pixels_fs),
                dtype=self._xp.float32,
            )
        ]

    def _setup_constant_buffers(self):
        assert (
            self._output_dtype == np.float32
        ), "Alternative output types not supported yet"

        self.offset_map = np.empty(self._map_shape, dtype=np.float32)
        self.rel_gain_map = np.empty(self._map_shape, dtype=np.float32)
        self.bad_pixel_map = np.empty(self._map_shape, dtype=np.uint32)
        self.noise_map = np.empty(self._map_shape, dtype=np.float32)
        # will do everything by quadrant
        self.thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=4)
        self._q_offset_map = utils.quadrant_views(self.offset_map)
        self._q_rel_gain_map = utils.quadrant_views(self.rel_gain_map)
        self._q_bad_pixel_map = utils.quadrant_views(self.bad_pixel_map)
        self._q_noise_map = utils.quadrant_views(self.noise_map)

    def _load_constant(self, constant_type, data):
        if constant_type is Constants.OffsetEPix100:
            self.offset_map[:] = data.squeeze().astype(np.float32)
        elif constant_type is Constants.RelativeGainEPix100:
            self.rel_gain_map[:] = data.squeeze().astype(np.float32)
        elif constant_type is Constants.BadPixelsDarkEPix100:
            self.bad_pixel_map[:] = (data.squeeze() & self.bad_pixel_subset)
        elif constant_type is Constants.NoiseEPix100:
            self.noise_map[:] = data.squeeze()
        else:
            raise Exception(f"Unknown constant type {constant_type}")

    def flush_buffers(self, constants):
        if Constants.OffsetEPix100 in constants:
            self.offset_map.fill(0)
        if Constants.RelativeGainEPix100 in constants:
            self.rel_gain_map.fill(1)
        if constants & {Constants.BadPixelsDarkEPix100}:
            self.bad_pixel_map.fill(0)
        if Constants.NoiseEPix100 in constants:
            self.noise_map.fill(np.inf)

    def _preview_data_views(self, raw_data, processed_data):
        # NOTE: apparently there are "calibration rows"
        # geometry assumes these are cut out already
        return (raw_data[2:-2], processed_data[2:-2])

    def _correct_quadrant(
        self,
        flags,
        input_data,
        offset_map,
        noise_map,
        bad_pixel_map,
        rel_gain_map,
        output,
    ):
        output[:] = input_data.astype(np.float32)
        if flags & CorrectionFlags.OFFSET:
            output -= offset_map

        if flags & CorrectionFlags.COMMONMODE:
            # per rectangular block that looks like something is going on
            cm_mask = (
                (bad_pixel_map != 0)
                | (output > noise_map * self._cm_sigma)
            ).astype(np.uint8, copy=False)
            masked = np.ma.masked_array(data=output, mask=cm_mask)
            if self._cm_block:
                for block in np.hsplit(masked, 4):
                    if block.count() < block.size * self._cm_minfrac:
                        continue
                    block.data[:] -= np.ma.median(block)

            if self._cm_row:
                common_cython.cm_fs(output, cm_mask, self._cm_sigma, self._cm_minfrac)

            if self._cm_col:
                common_cython.cm_ss(output, cm_mask, self._cm_sigma, self._cm_minfrac)

        if flags & CorrectionFlags.RELGAIN:
            output *= rel_gain_map

        if flags & CorrectionFlags.BPMASK:
            output[bad_pixel_map != 0] = self.bad_pixel_mask_value

    def _correct(self, flags, image_data, cell_table, output):
        # ignore cell table None
        concurrent.futures.wait(
            [
                self.thread_pool.submit(
                    self._correct_quadrant(flags, *parts)
                )
                for parts in zip(
                        utils.quadrant_views(image_data),
                        self._q_offset_map,
                        self._q_noise_map,
                        self._q_bad_pixel_map,
                        self._q_rel_gain_map,
                        utils.quadrant_views(output),
                )
            ]
        )

    def __del__(self):
        self.thread_pool.shutdown()


@KARABO_CLASSINFO("Epix100Correction", deviceVersion)
class Epix100Correction(base_correction.BaseCorrection):
    _base_output_schema = schemas.jf_output_schema
    _kernel_runner_class = Epix100CpuRunner
    _correction_steps = correction_steps
    _correction_flag_class = CorrectionFlags
    _constant_enum_class = Constants
    _calcat_friend_class = Epix100CalcatFriend

    _image_data_path = "data.image.pixels"
    _cell_table_path = None
    _pulse_table_path = None

    _preview_outputs = [
        PreviewSpec(
            name,
            dimensions=2,
            frame_reduction=False,
        )
        for name in ("raw", "corrected")
    ]
    _warn_memory_cell_range = False
    _use_shmem_handles = False

    @classmethod
    def expectedParameters(cls, expected):
        cls._calcat_friend_class.add_schema(expected)
        cls._kernel_runner_class.add_schema(expected)
        base_correction.add_addon_nodes(expected, cls)
        PreviewFriend.add_schema(expected, cls._preview_outputs)
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(
                cls._base_output_schema(use_shmem_handles=cls._use_shmem_handles)
            )
            .commit(),

            # TODO: disable preview selection mode

            OVERWRITE_ELEMENT(expected)
            .key("useShmemHandles")
            .setNewDefaultValue(cls._use_shmem_handles)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("outputShmemBufferSize")
            .setNewDefaultValue(2)
            .commit(),
        )

    def _get_data_from_hash(self, data_hash):
        image_data = data_hash.get(self._image_data_path)
        if self.unsafe_get("workarounds.overrideInputAxisOrder"):
            expected_shape = self.kernel_runner.expected_input_shape(1)
            if expected_shape != image_data.shape:
                image_data.shape = expected_shape
        return (
            1,
            image_data,
            None,
            None,
        )
