import enum

import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
)

from .. import (
    base_calcat,
    base_correction,
    base_kernel_runner,
    schemas,
)
from ..preview_utils import PreviewFriend, FrameSelectionMode, PreviewSpec
from .._version import version as deviceVersion


class Constants(enum.Enum):
    LUTGotthard2 = enum.auto()
    OffsetGotthard2 = enum.auto()
    RelativeGainGotthard2 = enum.auto()
    BadPixelsDarkGotthard2 = enum.auto()
    BadPixelsFFGotthard2 = enum.auto()


bad_pixel_constants = {
    Constants.BadPixelsDarkGotthard2,
    Constants.BadPixelsFFGotthard2,
}


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    LUT = 1
    OFFSET = 2
    GAIN = 4
    BPMASK = 8


correction_steps = (
    (
        "lut",
        CorrectionFlags.LUT,
        {Constants.LUTGotthard2},
    ),
    (
        "offset",
        CorrectionFlags.OFFSET,
        {Constants.OffsetGotthard2},
    ),
    (
        "gain",
        CorrectionFlags.GAIN,
        {Constants.RelativeGainGotthard2},
    ),
    (
        "badPixels",
        CorrectionFlags.BPMASK,
        {
            Constants.BadPixelsDarkGotthard2,
            Constants.BadPixelsFFGotthard2,
        }
    ),
)


def framesums_reduction_fun(data, cell_table, pulse_table, warn_fun):
    return np.nansum(data, axis=1)


class Gotthard2CpuRunner(base_kernel_runner.BaseKernelRunner):
    _bad_pixel_constants = bad_pixel_constants
    _correction_steps = correction_steps
    _correction_flag_class = CorrectionFlags
    _xp = np

    num_pixels_ss = None  # 1d detector, should never access this attribute
    num_pixels_fs = 1280

    @classmethod
    def add_schema(cls, schema):
        super(cls, cls).add_schema(schema)
        super(cls, cls).add_bad_pixel_config(schema)

    def expected_input_shape(self, num_frames):
        return (num_frames, self.num_pixels_fs)

    def _expected_output_shape(self, num_frames):
        return (num_frames, self.num_pixels_fs)

    def _make_output_buffers(self, num_frames, flags):
        return [
            self._xp.empty(
                (num_frames, self.num_pixels_fs), dtype=np.float32)
           ]

    def _preview_data_views(self, raw_data, raw_gain, processed_data):
        return [
            raw_data,  # raw 1d
            raw_gain,  # gain 1d
            processed_data,  # corrected 1d
            processed_data,  # frame sums
            raw_data,  # raw streak (2d)
            raw_gain,  # gain streak
            processed_data,  # corrected streak
        ]

    def _setup_constant_buffers(self):
        # model: 2 buffers (corresponding to actual memory cells), 2720 frames
        # lut maps from uint12 to uint10 values
        # TODO: override superclass properties instead
        self.lut_shape = (2, 4096, self.num_pixels_fs)
        self.map_shape = (3, self._constant_memory_cells, self.num_pixels_fs)
        self.lut = np.empty(self.lut_shape, dtype=np.uint16)
        self.offset_map = np.empty(self.map_shape, dtype=np.float32)
        self.rel_gain_map = np.empty(self.map_shape, dtype=np.float32)
        self.bad_pixel_map = np.empty(self.map_shape, dtype=np.uint32)
        self.flush_buffers(set(Constants))

    def _post_init(self):
        from ..kernels import gotthard2_cython
        self.correction_kernel = gotthard2_cython.correct

    def _load_constant(self, constant_type, data):
        if constant_type is Constants.LUTGotthard2:
            self.lut[:] = np.transpose(data.astype(np.uint16, copy=False), (1, 2, 0))
        elif constant_type is Constants.OffsetGotthard2:
            self.offset_map[:] = np.transpose(data.astype(np.float32, copy=False))
        elif constant_type is Constants.RelativeGainGotthard2:
            self.rel_gain_map[:] = np.transpose(data.astype(np.float32, copy=False))
        elif constant_type in bad_pixel_constants:
            # TODO: add the regular bad pixel subset configuration
            self.bad_pixel_map |= (np.transpose(data) & self.bad_pixel_subset)
        else:
            raise ValueError(f"What is this constant '{constant_type}'?")

    def flush_buffers(self, constants):
        if Constants.LUTGotthard2 in constants:
            default_lut = (
                np.arange(2 ** 12).astype(np.float64) * 2 ** 10 / 2 ** 12
            ).astype(np.uint16)
            self.lut[:] = np.stack(
                [np.stack([default_lut] * 2)] * self.num_pixels_fs, axis=2
            )
        if Constants.OffsetGotthard2 in constants:
            self.offset_map.fill(0)
        if Constants.RelativeGainGotthard2 in constants:
            self.rel_gain_map.fill(1)
        if bad_pixel_constants & constants:
            self.bad_pixel_map.fill(0)

    def _correct(self, flags, raw_data, cell_table, gain_stages, processed_data):
        self.correction_kernel(
            raw_data,
            gain_stages,
            flags,
            self.lut,
            self.offset_map,
            self.rel_gain_map,
            self.bad_pixel_map,
            self.bad_pixel_mask_value,
            processed_data,
        )


class Gotthard2CalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = Constants

    def __init__(self, device, *args, **kwargs):
        super().__init__(device, *args, **kwargs)
        self._constants_need_conditions = {c: self.condition for c in Constants}

    def condition(self):
        res = base_calcat.OperatingConditions()
        res["Exposure time"] = self._get_param("exposureTime")
        res["Exposure period"] = self._get_param("exposurePeriod")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Single photon"] = self._get_param("singlePhoton")
        res["Acquisition rate"] = self._get_param("acquisitionRate")
        return res

    @staticmethod
    def add_schema(schema):
        super(Gotthard2CalcatFriend, Gotthard2CalcatFriend).add_schema(
            schema, "Gotthard2-Type"
        )

        # set some defaults for common parameters
        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsX")
            .setNewDefaultValue(1280)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsY")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(2)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(200)
            .commit()
        )

        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.exposureTime")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.exposurePeriod")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.singlePhoton")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.acquisitionRate")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(1.1)
            .reconfigurable()
            .commit(),
        )

        base_calcat.add_status_schema_from_enum(
            schema, Constants
        )


@KARABO_CLASSINFO("Gotthard2Correction", deviceVersion)
class Gotthard2Correction(base_correction.BaseCorrection):
    _base_output_schema = schemas.jf_output_schema
    _kernel_runner_class = Gotthard2CpuRunner
    _calcat_friend_class = Gotthard2CalcatFriend
    _correction_steps = correction_steps
    _constant_enum_class = Constants

    _image_data_path = "data.adc"
    _cell_table_path = None
    _pulse_table_path = None

    _warn_memory_cell_range = False  # for now, receiver always writes 255
    _cuda_pin_buffers = False
    _use_shmem_handles = False

    _preview_outputs = [
        PreviewSpec(
            "raw",
            dimensions=1,
            frame_reduction=True,
            wrap_in_imagedata=False,
            valid_frame_selection_modes=[FrameSelectionMode.FRAME],
        ),
        PreviewSpec(
            "gain",
            dimensions=1,
            frame_reduction=True,
            wrap_in_imagedata=False,
            valid_frame_selection_modes=[FrameSelectionMode.FRAME],
        ),
        PreviewSpec(
            "corrected",
            dimensions=1,
            frame_reduction=True,
            wrap_in_imagedata=False,
            valid_frame_selection_modes=[FrameSelectionMode.FRAME],
        ),
        PreviewSpec(
            "framesums",
            dimensions=1,
            frame_reduction=False,
            frame_reduction_fun=framesums_reduction_fun,
            wrap_in_imagedata=False,
        ),
        PreviewSpec(
            "rawStreak",
            dimensions=2,
            frame_reduction=False,
            wrap_in_imagedata=True,
        ),
        PreviewSpec(
            "gainStreak",
            dimensions=2,
            frame_reduction=False,
            wrap_in_imagedata=True,
        ),
        PreviewSpec(
            "correctedStreak",
            dimensions=2,
            frame_reduction=False,
            wrap_in_imagedata=True,
        ),
    ]

    @classmethod
    def expectedParameters(cls, expected):
        cls._calcat_friend_class.add_schema(expected)
        cls._kernel_runner_class.add_schema(expected)
        PreviewFriend.add_schema(expected, cls._preview_outputs)
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(
                cls._base_output_schema(use_shmem_handles=cls._use_shmem_handles)
            )
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("outputShmemBufferSize")
            .setNewDefaultValue(2)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("useShmemHandles")
            .setNewDefaultValue(cls._use_shmem_handles)
            .commit(),
        )

    def _get_data_from_hash(self, data_hash):
        image_data = np.asarray(
            data_hash.get(self._image_data_path)
        ).astype(np.uint16, copy=False)
        gain_data = np.asarray(
            data_hash.get("data.gain")
        ).astype(np.uint8, copy=False)
        num_frames = len(data_hash.get("data.frameNumber"))
        if self.unsafe_get("workarounds.overrideInputAxisOrder"):
            expected_shape = self.kernel_runner.expected_input_shape(num_frames)
            image_data.shape = expected_shape
            gain_data.shape = expected_shape
        return (
            num_frames,
            image_data,
            None,  # no cell table
            None,  # no pulse table
            gain_data,
        )
