import numpy as np
import pytest
from calng.corrections.JungfrauCorrection import (
    Constants,
    JungfrauBaseRunner,
    JungfrauCalcatFriend,
    JungfrauCpuRunner,
    JungfrauGpuRunner,
)
from calng.utils import BadPixelValues
from karabo.bound import Hash

from common_setup import DummyCorrectionDevice, get_constant_from_file, maybe_get


raw_data = np.rint(np.random.random((16, 512, 1024)) * 5000).astype(np.uint16)
gain = np.random.choice(
    np.uint8([0, 1, 3]), size=raw_data.shape
)  # not testing gain value 2 (wrong) for now
gain_as_index = gain.copy()
gain_as_index[gain_as_index == 3] = 2
constants = {
    Constants.Offset10Hz: get_constant_from_file(
        "xfel/cal/jungfrau-type/jungfrau_m572/",
        "cal.1714364366.1478188.h5",
        "/Jungfrau_M572/Offset10Hz/0",
    ),
    Constants.BadPixelsDark10Hz: get_constant_from_file(
        "xfel/cal/jungfrau-type/jungfrau_m572/",
        "cal.1714364369.7054155.h5",
        "/Jungfrau_M572/BadPixelsDark10Hz/0",
    ),
    Constants.BadPixelsFF10Hz: get_constant_from_file(
        "xfel/cal/jungfrau-type/jungfrau_m572/",
        "cal.1711037632.1430287.h5",
        "/Jungfrau_M572/BadPixelsFF10Hz/0",
    ),
    Constants.RelativeGain10Hz: get_constant_from_file(
        "xfel/cal/jungfrau-type/jungfrau_m572/",
        "cal.1711036758.5906427.h5",
        "/Jungfrau_M572/RelativeGain10Hz/0",
    ),
}
cell_table = np.arange(16, dtype=np.uint8)
np.random.shuffle(cell_table)
cast_data = raw_data.astype(np.float32)
offset_corrected = cast_data - np.choose(
    gain_as_index,
    np.transpose(constants[Constants.Offset10Hz][:, :, cell_table], (3, 2, 0, 1)),
)
gain_corrected = offset_corrected / np.choose(
    gain_as_index, constants[Constants.RelativeGain10Hz][:, :, cell_table].T
)
masked_dark = gain_corrected.copy()
masked_dark[
    np.choose(
        gain_as_index,
        np.transpose(constants[Constants.BadPixelsDark10Hz], (3, 2, 0, 1))[
            :, cell_table
        ],
    )
    != 0
] = np.nan
masked_ff = masked_dark.copy()
masked_ff[
    np.choose(
        gain_as_index,
        np.transpose(constants[Constants.BadPixelsFF10Hz], (3, 2, 1, 0))[:, cell_table],
    )
    != 0
] = np.nan


@pytest.fixture(params=[JungfrauCpuRunner, JungfrauGpuRunner])
def kernel_runner(request):
    device = DummyCorrectionDevice(JungfrauBaseRunner, JungfrauCalcatFriend)
    runner = request.param(device)
    device.runner = runner
    # TODO: test masking aroudn ASIC seams
    # TODO: test single-frame mode
    device.reconfigure(
        Hash(
            "constantParameters.memoryCells",
            16,
        )
    )
    device.reconfigure(
        Hash(
            "corrections.badPixels.subsetToUse.NON_STANDARD_SIZE",
            False,
        )
    )
    return runner


def test_cast(kernel_runner):
    _, result, previews = kernel_runner.correct(raw_data, cell_table, gain)
    result = maybe_get(result)
    assert np.allclose(result, cast_data, equal_nan=True)


def test_correct(kernel_runner):
    for constant, constant_data in constants.items():
        kernel_runner.load_constant(constant, constant_data)

    _, result, previews = kernel_runner.correct(raw_data, cell_table, gain)
    result = maybe_get(result)
    assert np.allclose(result, masked_ff, equal_nan=True)
