import itertools

import numpy as np
import pytest
from calng import utils
from calng.corrections.PnccdCorrection import (
    Constants,
    PnccdCalcatFriend,
    PnccdCpuRunner,
)
from karabo.bound import Hash

from common_setup import DummyCorrectionDevice


@pytest.fixture
def kernel_runner():
    device = DummyCorrectionDevice(PnccdCpuRunner, PnccdCalcatFriend)
    runner = PnccdCpuRunner(device)
    device.runner = runner
    return runner


def slow_common_mode(
    data, bad_pixel_map, noise_map, cm_min_frac, cm_noise_sigma, cm_ss, cm_fs
):
    masked = np.ma.masked_array(
        data=data,
        mask=(bad_pixel_map != 0) | (data > noise_map * cm_noise_sigma),
    )

    if cm_fs:
        subset_fs = masked.count(axis=1) >= masked.shape[1] * cm_min_frac
        data[subset_fs] -= np.ma.median(masked[subset_fs], axis=1, keepdims=True)

    if cm_ss:
        subset_ss = masked.count(axis=0) >= masked.shape[0] * cm_min_frac
        data[:, subset_ss] -= np.ma.median(masked[:, subset_ss], axis=0)


def test_common_mode(kernel_runner):
    data_shape = (1024, 1024)
    data = (np.random.random(data_shape) * 1000).astype(np.uint16)
    noise_maps = [
        (np.random.random(data_shape) * noise_level).astype(np.float32)
        for noise_level in (0, 5, 100, 1000)
    ]
    bad_pixel_maps = [
        (np.random.random(data_shape) < (bad_pixel_percentage / 100)).astype(np.uint32)
        for bad_pixel_percentage in (0, 1, 20, 80, 100)
    ]
    kernel_runner._device.reconfigure(
        Hash(
            "corrections.commonMode.noiseSigma",
            10,
            "corrections.commonMode.minFrac",
            0.25,
            "corrections.commonMode.enableRow",
            True,
            "corrections.commonMode.enableCol",
            True,
            "corrections.badPixels.enable",
            False,
        )
    )
    for noise_map, bad_pixel_map in itertools.product(noise_maps, bad_pixel_maps):
        kernel_runner.load_constant(Constants.NoiseCCD, noise_map)
        kernel_runner.load_constant(Constants.BadPixelsDarkCCD, bad_pixel_map)
        _, processed_data, _ = kernel_runner.correct(data, None)

        slow_version = data.astype(np.float32, copy=True)
        for q_data, q_noise, q_bad_pixels in zip(
            utils.quadrant_views(slow_version),
            utils.quadrant_views(noise_map),
            utils.quadrant_views(bad_pixel_map),
        ):
            slow_common_mode(q_data, q_bad_pixels, q_noise, 0.25, 10, True, True)
        assert np.allclose(processed_data, slow_version, equal_nan=True)
